package de.tub.mac13.teamb.util.visual;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.Serializable;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import de.tub.mac13.teamb.ontology.graph.Edge;
import de.tub.mac13.teamb.ontology.graph.Graph;
import de.tub.mac13.teamb.ontology.graph.Vertex;

/**
 * Graph Layouter for the Multi Agent Programming Contest Mars Exploration
 * scenario. The layouter uses the MASSIM classes for generating a graph just
 * like the ones used in the contest and then uses a swarm-based layouter for
 * positioning the nodes. It also contains a small GUI for viewing the layout.
 *
 * This layouter, or a derivative thereof, could be used in the agent contest
 * monitor for layouting the currently explored graph.
 *
 * @author kuester
 */
public class GraphLayouter implements ActionListener,Serializable{

	private static final long serialVersionUID = -2457529400035249762L;
	
	/**
     * random number generator
     */
    final Random random = new Random();
    static final int gridSize = 31;
    /**
     * nodes generated by the graph generator
     */
    public final Vector<GraphNode> nodes;
    /**
     * edges generated by the graph generator
     */
    public final Vector<GraphEdge> edges;
    /**
     * adjacency lists for all the nodes
     */
    final Map<Integer, Set<Integer>> adjacency = new ConcurrentHashMap<Integer, Set<Integer>>();
    /**
     * current position of all the (known) nodes
     */
    public final Map<Integer, Point> positions = new ConcurrentHashMap<Integer, Point>();
    /**
     * known nodes, only relevant for {@link #initializeFewNodes(int)}
     */
    final Set<Integer> knownNodes = new HashSet<Integer>();
    /**
     * fringe nodes, only relevant for {@link #initializeFewNodes(int)}
     */
    final Set<Integer> fringe = new HashSet<Integer>();
    /**
     * optional node currently selected in GUI (will not be repositioned)
     */
    Integer selected = null;
    /**
     * Added for Dynamic Mode
     */
    public final GraphNode[] nodeLookup;
    private ActionListener list;
    private boolean locked = false, wait = false;
    ConcurrentLinkedQueue<ActionEvent> blockedEvents = new ConcurrentLinkedQueue<>();
    Graph g;
    public Point gravity;

    /**
     * Create a new random graph using the MASSIM GraphGenerator.
     *
     */
    public GraphLayouter(Graph g) {
        // create new random graph
        this.nodes = new Vector<GraphNode>(g.getVertexCount());
        this.edges = new Vector<>();
        nodeLookup = new GraphNode[g.getVertexCount()];
        g.add(this);
        this.g = g;

        // create adjacency sets
        updateAdjecency();
    }

    private void clear() {
        positions.clear();
        knownNodes.clear();
        fringe.clear();
        blockedEvents.clear();
    }

    /**
     * Initialize layout with positions calculated by GraphCreator
     */
    public synchronized void initializeOriginalPositions() {
        clear();
        synchronized (nodes) {
            for (GraphNode node : nodes) {
                positions.put(node.id, new Point(node.gridX, node.gridY));
            }
        }
        knownNodes.addAll(positions.keySet());
    }

    /**
     * Initialize layout with random locations
     */
    public synchronized void initializeRandomPositions() {
        clear();
        synchronized (nodes) {
            for (GraphNode node : nodes) {
                positions.put(node.id, new Point(random.nextInt(gridSize),
                        random.nextInt(gridSize)));
            }
        }
        knownNodes.addAll(positions.keySet());
    }

    /**
     * Initialize layout with random positions for just a few nodes; all other
     * nodes are not yet "explored" and have to be added using
     * {@link #addNode()}.
     *
     * @param numNodes Number of nodes initially visible
     */
    public synchronized void initializeFewNodes(int numNodes) {
        clear();
        for (int i = 0; i < numNodes; i++) {
            int index = random.nextInt(nodes.size());
            Integer node = nodes.get(index).id;
            positions.put(
                    node,
                    new Point(random.nextInt(gridSize), random
                    .nextInt(gridSize)));
            knownNodes.add(node);
            fringe.addAll(adjacency.get(node));
        }
    }

    /**
     * Add another (random) node from the fringe set to the set of known nodes
     * and update the fringe set accordingly.
     */
    public synchronized void addNode() {
        for (Iterator<Integer> iter = fringe.iterator(); iter.hasNext();) {
            Integer next = iter.next();
            if (!knownNodes.contains(next)) {
                fringe.addAll(adjacency.get(next));
                knownNodes.add(next);
                positions.put(next, new Point(20, 20));
                break;
            }
            iter.remove();
        }
        System.out.println(knownNodes.size());
    }

    public synchronized void addNode(Vertex v) {
        int gridX = random.nextInt(gridSize + 1);
        int gridY = random.nextInt(gridSize + 1);
        int x = gridX * 75;
        int y = gridY * 75;

        GraphNode gn = new GraphNode(v.id, v.value, gridX, gridY, x, y);

        nodes.add(gn);
        nodeLookup[v.id] = gn;
        positions.put(v.id,
                new Point(random.nextInt(gridSize), random.nextInt(gridSize)));
        knownNodes.add(v.id);
    }

    public synchronized void updateNode(Vertex v) {
        if (v == null) {
            return;
        }
        GraphNode gn = nodeLookup[v.id];
        if (gn == null) {
            return;
        }
        gn.setDominatorTeam(v.getTeam());
        if(gn.weight < v.value)
            gn.weight = v.value;
    }

    public synchronized void addEdge(Edge e) {
        if (nodeLookup[e.node1] == null) {
            addNode(new Vertex(e.node1));
        }
        if (nodeLookup[e.node2] == null) {
            addNode(new Vertex(e.node2));
        }
        GraphEdge edge = new GraphEdge(1, nodeLookup[e.node1], nodeLookup[e.node2]);
        edges.add(edge);
        updateAdjecency();
    }

    private synchronized void updateAdjecency() {
        synchronized (edges) {
            for (GraphEdge edge : edges) {
                if (!adjacency.keySet().contains(edge.node1.id)) {
                    adjacency.put(edge.node1.id, new HashSet<Integer>());
                }
                if (!adjacency.keySet().contains(edge.node2.id)) {
                    adjacency.put(edge.node2.id, new HashSet<Integer>());
                }
                adjacency.get(edge.node1.id).add(edge.node2.id);
                adjacency.get(edge.node2.id).add(edge.node1.id);
            }
        }
    }

    /**
     * Calculate the "remoteliness" of all the nodes, by calculating the sum of
     * the distances to all the other nodes and normalizing it. The result if a
     * number between 0 and 1 for each node in the graph. This can be used to
     * determine the "rim" of the graph.
     *
     * @return mapping "node name" -> "remoteliness"
     */
    public synchronized Map<Integer, Double> getRemoteliness() {
        Map<Integer, Double> result = new HashMap<Integer, Double>();

        // get distances to all the other nodes
        double max = Double.MIN_VALUE;
        double min = Double.MAX_VALUE;
        for (Integer node : positions.keySet()) {
            Point p = positions.get(node);
            double d = 0;
            for (Integer other : positions.keySet()) {
                Point q = positions.get(other);
                d += Math.sqrt(Math.pow(p.x - q.x, 2) + Math.pow(p.y - q.y, 2));
            }
            result.put(node, d);
            max = Math.max(max, d);
            min = Math.min(min, d);
        }

        // normalize
        for (Integer node : positions.keySet()) {
            double d = result.get(node);
            double v = (d - min) / (max - min);
            result.put(node, v);
        }
        return result;
    }

    /**
     * Perform one iteration of Swarm-based layout. The basic idea is that each
     * node is (up to a certain limit) pulled towards each node it is connected
     * with, and at the same time (again, up to a certain limit) pushed away
     * from each node it is _not_ connected to.
     *
     * This method might still deserve some fine-tuning.
     *
     * @param calculateDistances calculate distances in each step, and use them
     * in the calculation
     */
    public synchronized double swarm(boolean calculateDistances) {
        Map<Integer, Double> distances = calculateDistances ? getRemoteliness()
                : null;
        Map<Integer, Point> backup = new ConcurrentHashMap<Integer, Point>(
                positions);

        // for each point in the current layout...
        double totalMovement = 0;
        for (Integer node : knownNodes) {
            if (node == selected) {
                continue;
            }

            Point p = backup.get(node);
            double moveX = 0;
            double moveY = 0;
            double dd = calculateDistances ? 1.5 - distances.get(node) : 1;

            // get distance and "delta" to each other node
            for (Integer other : backup.keySet()) {
                Point q = backup.get(other);
                double dx = q.x - p.x;
                double dy = q.y - p.y;
                double d = Math.sqrt(dx * dx + dy * dy);

                // if there is an edge between the nodes, move towards the
                // node, otherwise move away from the node
                if (d > 0) {
                    double f = 0;
                    if (adjacency.get(node) == null) {
                        continue;
                    }
                    if (adjacency.get(node).contains(other)) {
                        if (d > 2) {
                            f = +0.01 * d * dd;
                        }
                    } else {
                        if (d < 4) {
                            f = -0.01 / d;
                        }
                    }
                    moveX += f * dx / d;
                    moveY += f * dy / d;
                }
            }
            // move the node in the new layout
            positions.put(node, new Point(p.x + moveX, p.y + moveY));
            totalMovement += Math.sqrt(moveX * moveX + moveY * moveY);
        }
        return totalMovement;
    }

    public void gravitate() {
        for(Integer i:positions.keySet()){
            Point p = positions.get(i);
            double dx = gravity.x-p.x;
            double dy = gravity.y-p.y;

            double d = Math.sqrt(dx * dx + dy * dy);
                   d *= 0.1;
                   d = Math.round(d);

            p.x += Math.signum(dx)*d;
            p.y += Math.signum(dy)*d;
            positions.put(i,p);
        }
    }

    /**
     * Simple class for a point.
     *
     * @author kuester
     */
    public static class Point implements Serializable{
    	
		private static final long serialVersionUID = 8313550966642045810L;
		
		public double x;
        public double y;

        public Point(double x, double y) {
            this.x = x;
            this.y = y;
        }
    }

    public void addActionListener(ActionListener list) {
        this.list = list;
    }

    public synchronized void unlock() {
        wait = true;
        locked = false;
        while (!blockedEvents.isEmpty()) {
            actionPerformed(blockedEvents.poll());
        }
        wait = false;

        if (list != null) {
            list.actionPerformed(new ActionEvent(this, 1, null));
        }
    }

    public synchronized void lock() {
        while (!blockedEvents.isEmpty()) {
            actionPerformed(blockedEvents.poll());
        }
        locked = true;
        wait = false;
    }

    @Override
    public void actionPerformed(ActionEvent e) {

        switch (e.getID()) {
            case Graph.ADDED_VERT:
                addNode((Vertex) e.getSource());
                break;
            case Graph.ADDED_EDGE:
                if (locked) {
                    blockedEvents.offer(e);
                } else {
                    addEdge((Edge) e.getSource());
                }
                break;
            case Graph.UPDATED_VERT:
                updateNode((Vertex) e.getSource());
                break;
            default:
                break;
        }

        if (!locked && list != null && !wait) {
            list.actionPerformed(new ActionEvent(this, 0, null));
        }

    }
}
